#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <alloca.h>
#include <err.h>

typedef unsigned char byte;
typedef unsigned int uint;
typedef unsigned short ushort;

struct top {
	uint len;
	void* buf;

	int fd;

	int cdsize;
	int cdsoff;

	void* cdbuf;
	uint cdlen;
	uint cdptr;
	uint ncdr;

	uint total;
} ctx;

struct mbuf {
	void* buf;
	uint len;
};

struct eocd {
	byte sig[4];
	ushort disk;
	ushort dcds;
	ushort ncdr;
	ushort tncd;
	uint cdsize;
	uint cdsoff;
	ushort clen;
	char comm[];
} __attribute__((packed));

struct lfrd {
	byte sig[4];
	ushort xver;
	ushort flags;
	ushort method;
	ushort mtime;
	ushort mdate;
	uint crc;
	uint compressed;
	uint uncompressed;
	ushort n;
	ushort m;
	byte name[];
} __attribute__((packed));

struct cdfr {
	byte sig[4];
	ushort cver;
	ushort xver;
	ushort flags;
	ushort method;
	ushort mtime;
	ushort mdate;
	uint crc;
	uint compressed;
	uint uncompressed;
	ushort n;
	ushort m;
	ushort k;
	ushort disk;
	ushort iattrs;
	uint eattrs;
	uint off;
	byte name[];
} __attribute__((packed));

static void mmap_whole_file(char* name)
{
	int fd;
	struct stat st;

	if((fd = open(name, O_RDONLY)) < 0)
		err(-1, "open");
	if(fstat(fd, &st) < 0)
		err(-1, "stat");

	uint size = st.st_size;
	void* buf;

	if(size != st.st_size)
		errx(-1, "file too large");

	if((buf = mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0)) == MAP_FAILED)
		err(-1, "mmap");

	if(close(fd) < 0)
		err(-1, "close");

	ctx.len = size;
	ctx.buf = buf;
}

static void locate_central_directory(void)
{
	uint len = ctx.len;
	void* buf = ctx.buf;
	struct eocd* eocd;

	if(len < sizeof(*eocd))
		errx(-1, "file too short");

	uint off = len - sizeof(*eocd);

	while(off > 0) {
		eocd = buf + off;

		if(memcmp(eocd->sig, "PK\x05\x06", 4)) {
			off--;
			continue;
		}

		goto got;
	};

	errx(-1, "cannot locate EOCD");
got:;
	uint cdsize = eocd->cdsize;
	uint cdsoff = eocd->cdsoff;

	if(cdsoff >= len)
		errx(-1, "invalid cdsoff");
	if(cdsoff + cdsize > len)
		errx(-1, "invalid cdsize");

	ctx.cdsize = cdsize;
	ctx.cdsoff = cdsoff;

	ctx.ncdr = eocd->ncdr;

}

static void copy_central_directory(void)
{
	uint size = ctx.cdsize;
	uint need = size + 2*4096;

	need = (need + 4095) & ~0xFFF;

	int prot = PROT_READ | PROT_WRITE;
	int flags = MAP_PRIVATE | MAP_ANONYMOUS;

	void* buf = mmap(NULL, need, prot, flags, -1, 0);

	if(buf == MAP_FAILED)
		err(-1, "mmap");

	memcpy(buf, ctx.buf + ctx.cdsoff, size);

	ctx.cdbuf = buf;
	ctx.cdlen = need;
	ctx.cdptr = size;
}

//static void dump_central_directory(void)
//{
//	void* ptr = ctx.cdbuf;
//	void* end = ptr + ctx.cdptr;
//
//	while(ptr < end) {
//		struct cdfr* rec = ptr;
//
//		if(memcmp(rec->sig, "PK\x01\x02", 4))
//			errx(-1, "invalid file record");
//
//		ptr += sizeof(*rec);
//		ptr += rec->n;
//		ptr += rec->m;
//		ptr += rec->k;
//
//		printf("REC %.*s\n", rec->n, rec->name);
//	}
//}

static void open_output_file(char* name)
{
	int fd;
	int flags = O_WRONLY | O_CREAT | O_TRUNC;
	int mode = 0644;

	if((fd = open(name, flags, mode)) < 0)
		err(-1, "create");

	ctx.fd = fd;
}

static void write_bulk_data(void)
{
	int ret, fd = ctx.fd;

	void* buf = ctx.buf;
	uint size = ctx.cdsoff;

	if((ret = write(fd, buf, size)) < 0)
		err(-1, "write");
	if(ret < size)
		errx(-1, "incomplete write");

	ctx.total = size;
}

static void* alloc(uint size)
{
	if(ctx.cdptr + size > ctx.cdlen)
		errx(-1, "out of CD space");

	void* ptr = ctx.cdbuf + ctx.cdptr;

	ctx.cdptr += size;

	return ptr;
}

static void append_end_of_directory(void)
{
	uint cdptr = ctx.cdptr;

	struct eocd* cd = alloc(sizeof(*cd));

	memcpy(cd->sig, "PK\x05\x06", 4);

	//printf("ncdr=%04X\n", ctx.ncdr);

	cd->disk = 0;
	cd->dcds = 0;
	cd->ncdr = ctx.ncdr;
	cd->tncd = ctx.ncdr;
	cd->cdsize = cdptr;
	cd->cdsoff = ctx.total;
	cd->clen = 0;
}

static void write_central_directory(void)
{
	int ret, fd = ctx.fd;
	uint len = ctx.cdptr;
	void* buf = ctx.cdbuf;

	if((ret = write(fd, buf, len)) < 0)
		err(-1, "write");
	if(ret != len)
		errx(-1, "incomplete write");
}

static void load_file(struct mbuf* mb, char* path)
{
	int fd, size;
	struct stat st;

	if((fd = open(path, O_RDONLY)) < 0)
		err(-1, "open %s", path);
	if(fstat(fd, &st) < 0)
		err(-1, "stat %s", path);

	size = st.st_size;

	void* buf;
	int prot = PROT_READ;
	int flags = MAP_PRIVATE;

	if((buf = mmap(NULL, size, prot, flags, fd, 0)) == MAP_FAILED)
		err(-1, "mmap %s", path);

	if(close(fd) < 0)
		err(-1, "close %s", path);

	mb->buf = buf;
	mb->len = size;
}

static void append_fdata(struct mbuf* mb, char* name)
{
	uint len = strlen(name);
	uint need = sizeof(struct lfrd) + len;
	struct lfrd* rec = alloca(need);

	memset(rec, 0, sizeof(*rec));
	memcpy(rec->sig, "PK\x03\x04", 4);
	rec->uncompressed = mb->len;
	rec->compressed = mb->len;
	rec->n = len;
	memcpy(rec->name, name, len);

	if(write(ctx.fd, rec, need) < 0)
		err(-1, "write");
	if(write(ctx.fd, mb->buf, mb->len) < 0)
		err(-1, "write");

	ctx.total += need + mb->len;
}

static void append_entry(struct mbuf* mb, char* name)
{
	uint len = strlen(name);
	uint need = sizeof(struct cdfr) + len;
	struct cdfr* rec = alloc(need);

	memset(rec, 0, sizeof(*rec));
	memcpy(rec->sig, "PK\x01\x02", 4);
	rec->uncompressed = mb->len;
	rec->compressed = mb->len;
	rec->n = len;
	memcpy(rec->name, name, len);

	ctx.ncdr++;
}

static void unmap_buf(struct mbuf* mb)
{
	if(munmap(mb->buf, mb->len) < 0)
		err(-1, "unmap");

	mb->buf = NULL;
	mb->len = 0;
}

static void append_file(char* path, char* name)
{
	struct mbuf mb;

	load_file(&mb, path);
	append_fdata(&mb, name);
	append_entry(&mb, name);
	unmap_buf(&mb);
}

int main(int argc, char** argv)
{
	mmap_whole_file("M3K.fw.orig");
	locate_central_directory();
	copy_central_directory();
	//dump_central_directory();

	open_output_file("M3K.fw");
	write_bulk_data();

	append_file("script.sh", "/etc/init.d/S00.sh");

	append_end_of_directory();
	write_central_directory();

	return 0;
}
